package com.jclark.xsl;

class AttributeListTemplateImpl implements AttributeListTemplate {
  private static
  class AttributeListImpl implements AttributeList {
    private AttributeList atts;
    private String[] values;
    AttributeListImpl(AttributeList atts, String[] values) {
      this.atts = atts;
      this.values = values;
    }
    public int getAttributeCount() {
      return atts.getAttributeCount();
    }
    public int getAttributeIndex(Name name) {
      return atts.getAttributeIndex(name);
    }
    public String getAttributeValue(Name name) {
      int i = atts.getAttributeIndex(name);
      if (i < 0)
	return null;
      return values[i];
    }
    public String getAttributeValue(int i) {
      return values[i];
    }
    public Name getAttributeName(int i) {
      return atts.getAttributeName(i);
    }
  }

  private static
  class UnquoteAttributeList implements AttributeList {
    private AttributeList atts;
    private Name[] names;
    UnquoteAttributeList(AttributeList atts, Name[] names) {
      this.atts = atts;
      this.names = names;
    }
    public int getAttributeCount() {
      return names.length;
    }
    public int getAttributeIndex(Name name) {
      for (int i = 0; i < names.length; i++)
	if (names[i].equals(name))
	  return i;
      return -1;
    }
    public String getAttributeValue(Name name) {
      int i = getAttributeIndex(name);
      if (i < 0)
	return null;
      return atts.getAttributeValue(i);
    }
    public String getAttributeValue(int i) {
      return atts.getAttributeValue(i);
    }
    public Name getAttributeName(int i) {
      return names[i];
    }
  }

  private static
  class InsertAttributeList implements AttributeList {
    private AttributeList atts;
    private int replaceIndex;
    private AttributeList insertAtts;
    // Attributes in the inserted list may be duplicates of attributes in the
    // list into which the attributes are being inserted.
    // Each non-duplicate attribute has a new index giving its index
    // in the inserted list with duplicates removed.
    // newIndex[i] gives the new index of the attribute with original index i
    private int[] newIndex;
    // oldIndex[i] gives the old index of the attribute with new index i
    private int[] oldIndex;
    InsertAttributeList(AttributeList atts, int replaceIndex, AttributeList insertAtts) {
      this.atts = atts;
      this.replaceIndex = replaceIndex;
      this.insertAtts = insertAtts;
      newIndex = new int[insertAtts == null ? 0 : insertAtts.getAttributeCount()];
      int n = 0;
      for (int i = 0; i < newIndex.length; i++)
	if (atts.getAttributeIndex(insertAtts.getAttributeName(i)) < 0)
	  newIndex[i] = n++;
      oldIndex = new int[n];
      n = 0;
      for (int i = 0; i < oldIndex.length; i++)
	if (atts.getAttributeIndex(insertAtts.getAttributeName(i)) < 0)
          oldIndex[n++] = i;
    }
    public int getAttributeCount() {
      return atts.getAttributeCount() - 1 + oldIndex.length;
    }
    public int getAttributeIndex(Name name) {
      int i = atts.getAttributeIndex(name);
      if (i < 0) {
	i = insertAtts.getAttributeIndex(name);
	if (i >= 0)
	  return newIndex[i];
      }
      else if (i >= replaceIndex)
	i += oldIndex.length - 1;
      return i;
    }
    public String getAttributeValue(Name name) {
      String value = atts.getAttributeValue(name);
      if (value != null)
	return value;
      return insertAtts.getAttributeValue(name);
    }
    public String getAttributeValue(int i) {
      if (i >= replaceIndex) {
	if (i < replaceIndex + oldIndex.length)
	  return insertAtts.getAttributeValue(oldIndex[i - replaceIndex]);
	i -= oldIndex.length;
      }
      return atts.getAttributeValue(i);
    }
    public Name getAttributeName(int i) {
      if (i >= replaceIndex) {
	if (i < replaceIndex + oldIndex.length)
	  return insertAtts.getAttributeName(oldIndex[i - replaceIndex]);
	i -= oldIndex.length;
      }
      return atts.getAttributeName(i);
    }
  }

  private Node node;
  private Expression[] valueTemplates;
  private Name[] names;
  private int useIndex = -1;
  private Name useName;

  AttributeListTemplateImpl(Node node) throws XSLException {
    this.node = node;
    valueTemplates = null;
    names = null;
    int len = node.getAttributeCount();
    for (int i = 0; i < len; i++) {
      Name name = node.getAttributeName(i);
      if (SheetImpl.XSL_NAMESPACE.equals(name.getNamespace())
	  && "use".equals(name.getLocalPart())) {
	useIndex = i;
	useName = node.getNamespacePrefixMap().expandElementTypeName(node.getAttributeValue(i), node);
      }
      else {
	String value = node.getAttributeValue(i);
	if (value.indexOf('{') >= 0 || value.indexOf('}') >= 0) {
	  if (valueTemplates == null) {
	    valueTemplates = new Expression[len];
	    for (int j = 0; j < i; j++)
	      valueTemplates[j] = new ConstantExpression(node.getAttributeValue(j));
	  }
	  valueTemplates[i] = PatternParser.createAttributeValueTemplate(node, value);
	}
	else if (valueTemplates != null)
	  valueTemplates[i] = new ConstantExpression(value);
	Name unquotedName = SheetImpl.unquoteName(name, node);
	if (unquotedName != name) {
	  if (names == null) {
	    names = new Name[len];
	    for (int j = 0; j < i; j++)
	      names[j] = node.getAttributeName(j);
	  }
	  names[i] = unquotedName;
	}
	else if (names != null)
	  names[i] = unquotedName;
      }
    }
  }

  public
  AttributeList instantiate(ProcessContext context, Node sourceNode) throws XSLException {
    AttributeList result = node;
    if (names != null)
      result = new UnquoteAttributeList(result, names);
    if (valueTemplates != null) {
      String[] values = new String[valueTemplates.length];
      for (int i = 0; i < values.length; i++) {
	if (i != useIndex)
	  values[i] = valueTemplates[i].eval(context, sourceNode);
      }
      result = new AttributeListImpl(result, values);
    }
    if (useIndex >= 0) {
      AttributeListTemplate insert = context.getNamedAttributeSet(useName);
      if (insert == null)
	result = new InsertAttributeList(result, useIndex, null);
      else
	result = new InsertAttributeList(result,
					 useIndex,
					 insert.instantiate(context, node));
    }
    return result;
  }
}

